'''
BSD Licence
Copyright (c) 2015, Science & Technology Facilities Council (STFC)
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
        this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
        this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.
    * Neither the name of the Science & Technology Facilities Council (STFC)
        nor the names of its contributors may be used to endorse or promote
        products derived from this software without specific prior written
        permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


Contents:
This module contains views of the user registration pages.

Created on 11 Dec 2013

@author: mnagni
'''
from json import dumps
import logging

from django.contrib.auth.models import User
from django.contrib.auth.views import login
from django.core.urlresolvers import reverse
from django.db.models import ObjectDoesNotExist
from django.db.utils import IntegrityError
from django.forms.util import ErrorList
from django.http.response import HttpResponseRedirect, HttpResponse, \
    HttpResponseNotFound
from django.http.response import HttpResponseServerError
from django.utils.decorators import method_decorator
from django.views.generic import TemplateView, View
from provider.oauth2.models import AccessToken

from djcharme import mm_render_to_response
from djcharme.charme_security_model import UserForm, UserUpdateForm, \
    UserProfileUpdateForm
from djcharme.forms import UsernameReminderForm
from djcharme.models import UserProfile
from djcharme.node import get_users_admin_role_orgs
from djcharme.security_middleware import is_valid_token
from djcharme.settings import REDIRECT_FIELD_NAME
from djcharme.views import get_extra_context, get_safe_redirect
from djcharme.views import not_authenticated


LOGGING = logging.getLogger(__name__)

CHANGE_FORM_TEMPLATE = 'registration/profile_change_form.html'
PROFILE_TEMPLATE = 'registration/profile.html'
REGISTEATION_TEMPLATE = 'registration/registration.html'


class Registration(View):
    """
    Display the registration view.

    """
    @method_decorator(not_authenticated)
    def dispatch(self, *args, **kwargs):
        try:
            return super(Registration, self).dispatch(*args, **kwargs)
        except Exception as ex:
            LOGGING.error("Registration - unexpected error: %s", ex)
            return HttpResponseServerError(str(ex))

    def get(self, request, *args, **kwargs):
        redirect_to = get_safe_redirect(request)
        context = {
            'user_form': UserForm(),
            'openid': False,
            REDIRECT_FIELD_NAME: redirect_to,
        }
        return mm_render_to_response(request, context, REGISTEATION_TEMPLATE)

    def post(self, request, *args, **kwargs):
        LOGGING.debug('Registration request received')
        redirect_to = get_safe_redirect(request)
        user_form = UserForm(request.POST)
        if user_form.is_valid():
            try:
                user = User.objects.create_user(
                    user_form.cleaned_data.get('username'),
                    user_form.cleaned_data.get('email'),
                    password=user_form.cleaned_data.get('password'),
                    first_name=user_form.cleaned_data.get('first_name'),
                    last_name=user_form.cleaned_data.get('last_name'))
                user.save()
                user_profile = UserProfile.objects.create(
                    user_id=user.id,
                    show_email=user_form.cleaned_data.get('show_email'))
                user_profile.save()
                login(request, user)
                return HttpResponseRedirect(redirect_to)

            except IntegrityError:
                LOGGING.debug('Username is already registered')
                errors = user_form._errors.setdefault('username', ErrorList())
                errors.append(u'Username is already registered')

        # form not valid
        context = {
            'user_form': user_form,
            'openid': False,
            REDIRECT_FIELD_NAME: redirect_to,
        }
        extra_context = get_extra_context(kwargs)
        if extra_context is not None:
            context.update(extra_context)
        return mm_render_to_response(request, context, REGISTEATION_TEMPLATE)


class Profile(TemplateView):
    """
    Display the profile view.

    """
    template_name = PROFILE_TEMPLATE

    def get_context_data(self, **kwargs):
        context = super(Profile, self).get_context_data(**kwargs)
        return context


def _update_user(request):
    LOGGING.debug('_update_user')
    context = {}
    user_form = UserUpdateForm(request.POST, instance=request.user)
    create_profile = False
    try:
        user_profile_form = UserProfileUpdateForm(
            request.POST, instance=request.user.userprofile)
    except ObjectDoesNotExist:
        user_profile_form = UserProfileUpdateForm(request.POST)
        create_profile = True
    if user_form.is_valid() and user_profile_form.is_valid:
        try:
            user_form.save()
            if create_profile:
                user_profile = UserProfile.objects.create(
                    user_id=request.user.id,
                    show_email=user_profile_form.base_fields.get('show_email'))
                user_profile.save()
            else:
                user_profile_form.save()
            context = {}
            context['msg'] = 'Profile updated'
            return mm_render_to_response(request, context, PROFILE_TEMPLATE)
        except IntegrityError:
            LOGGING.debug('Username is already registered')
            errors = user_form._errors.setdefault('username', ErrorList())
            errors.append(u'Username is already registered')

    context['user_form'] = user_form
    context['user_profile_form'] = user_profile_form
    return mm_render_to_response(request, context, CHANGE_FORM_TEMPLATE)


def profile_change(request):
    context = {}
    LOGGING.debug('Profile request received')

    if request.method == 'POST':
        LOGGING.debug('method is POST')
        return _update_user(request)
    else:  # GET
        # Set up initial values
        user = request.user
        orig_values = {}
        orig_values['username'] = user.get_username
        orig_values['first_name'] = user.first_name
        orig_values['last_name'] = user.last_name
        orig_values['email'] = user.email
        user_form = UserUpdateForm(initial=orig_values)
        context['user_form'] = user_form

        orig_values = {}
        try:
            user_profile = user.userprofile
            orig_values['show_email'] = user_profile.show_email
        except ObjectDoesNotExist:
            orig_values['show_email'] = False
        user_profile_form = UserProfileUpdateForm(initial=orig_values)
        context['user_profile_form'] = user_profile_form
        return mm_render_to_response(request, context, CHANGE_FORM_TEMPLATE)


def username_reminder(request, from_email=None):
    username_reminder_form = UsernameReminderForm
    post_reset_redirect = reverse('username_reminder_done')
    if request.method == "POST":
        form = username_reminder_form(request.POST)
        if form.is_valid():
            opts = {
                'use_https': request.is_secure(),
                'from_email': from_email,
                'email_template_name':
                    'registration/username_reminder_email.html',
                'subject_template_name':
                    'registration/username_reminder_subject.txt',
                'request': request,
            }
            form.save(**opts)
            return HttpResponseRedirect(post_reset_redirect)
    else:
        form = username_reminder_form()
    context = {
        'form': form,
    }

    return mm_render_to_response(request, context,
                                 'registration/username_reminder_form.html')


def username_reminder_done(request):
    context = {}
    return mm_render_to_response(request, context,
                                 'registration/username_reminder_done.html')


def validate_token(request, token=None, expire=None):
    if is_valid_token(token):
        return HttpResponse(status=200)
    return HttpResponseNotFound()


def userinfo(request):
    # The request has an Access Token
    if request.environ.get('HTTP_AUTHORIZATION', None):
        for term in request.environ.get('HTTP_AUTHORIZATION').split():
            try:
                access_t = AccessToken.objects.get(token=term)
                ret = {}
                ret['username'] = access_t.user.username
                ret['first_name'] = access_t.user.first_name
                ret['last_name'] = access_t.user.last_name
                ret['admin_for'] = ','.join(get_users_admin_role_orgs
                                            (request.user.id))
                return HttpResponse(dumps(ret),
                                    content_type="application/json")
            except AccessToken.DoesNotExist:
                continue
    return HttpResponseNotFound()


def token_response(request):
    return mm_render_to_response(request, {}, 'token_response.html')


def test_token(request):
    return mm_render_to_response(request, {}, 'oauth_test2.html')




'''
BSD Licence
Copyright (c) 2014, Science & Technology Facilities Council (STFC)
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
        this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
        this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.
    * Neither the name of the Science & Technology Facilities Council (STFC)
        nor the names of its contributors may be used to endorse or promote
        products derived from this software without specific prior written
        permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Created on 9 Jan 2012

@author: Maurizio Nagni
'''


import logging
import mimetypes
from multiprocessing.process import Process

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.models import User
from django.db.utils import DatabaseError
from django.http.response import HttpResponse
from rdflib.plugins.stores.sparqlstore import SPARQLUpdateStore

from djcharme import mm_render_to_response_error, LOAD_SAMPLE


def webusage(request):
    template = ('\nMETHOD:%s\nIP:%s\nREMOTE_HOST:%s\nPATH_INFO:%s\n' \
                'HTTP_USER_AGENT:%s\n')
    return template % (request.META.get('REQUEST_METHOD', 'Unknown'),
                       request.META.get('REMOTE_ADDR', 'Unknown'),
                       request.META.get('REMOTE_HOST', 'Unknown'),
                       request.META.get('PATH_INFO', 'Unknown'),
                       request.META.get('HTTP_USER_AGENT', 'Unknown'))


LOGGING = logging.getLogger(__name__)

USAGE_LOG = logging.getLogger('webusage')
USAGE_LOG.setLevel(logging.DEBUG)
# create console handler with a higher log level
HANDLER = logging.StreamHandler()
HANDLER.setLevel(logging.INFO)

# create file handler which logs even debug messages
# ch = logging.FileHandler('spam.log')
# ch.setLevel(logging.INFO)

# create formatter and add it to the handlers
FORMATTER = logging.Formatter(fmt='%(name)s %(asctime)s \
%(message)s')
HANDLER.setFormatter(FORMATTER)
USAGE_LOG.addHandler(HANDLER)

# USAGE_LOG.basicConfig(format='%(name)s:%(levelname)s:%(message)s',
# level=logging.INFO,datefmt='%d/%m/%y %I:%M:%S')

if not mimetypes.inited:
    mimetypes.init()
    mimetypes.add_type('application/rdf+xml', '.rdf')
    mimetypes.add_type('text/turtle', '.ttl')
    mimetypes.add_type('application/ld+json', '.json-ld')


class CharmeMiddleware(object):

    __store = None
    __osEngine = None

    DEFAULT_OPTIONS_HDR_RESPONSE = {
        'Access-Control-Allow-Methods': 'GET, OPTIONS, POST, DELETE',
        'Access-Control-Allow-Headers': 'X-CSRFToken, X-Requested-With, ' \
            'x-requested-with, Content-Type, Content-Length, Authorization',
        'Access-Control-Max-Age': 10,
        'Content-Type': "text/plain"
    }

    @classmethod
    def __init_os_engine(self):
        from djcharme.opensearch.os_conf import setUp
        LOGGING.info("__init_os_engine - OpenSearch Engine created")
        CharmeMiddleware.__osEngine = setUp()

    @classmethod
    def __init_store(self):
        store = SPARQLUpdateStore(queryEndpoint=getattr(settings,
                                                        'SPARQL_QUERY'),
                                  update_endpoint=getattr(settings,
                                                          'SPARQL_UPDATE'),
                                 postAsEncoded=False)
        store.bind("chnode", getattr(settings, 'NODE_URI', 'http://localhost'))
        LOGGING.info("__init_store - Store created")
        CharmeMiddleware.__store = store

        # Creates a superuser if there is not any
        try:
            users = User.objects.all()
            if len(users) == 0:
                User.objects.create_superuser('admin', '', 'admin')
        except DatabaseError:
            LOGGING.error("__init_store - Cannot find or create an "
                          "application superuser")

    @classmethod
    def get_store(self, debug=False):
        if debug or CharmeMiddleware.__store is None:
            CharmeMiddleware.__init_store()
            if getattr(settings, LOAD_SAMPLE, False):
                LOGGING.info("get_store - LOAD_SAMPLE: %s", True)
                from djcharme.node.sample import load_sample
                proc = Process(target=load_sample)  # inits thread
                proc.start()  # starts thread
        return CharmeMiddleware.__store

    @classmethod
    def get_osengine(self, debug=False):
        if debug or CharmeMiddleware.__osEngine is None:
            CharmeMiddleware.__init_os_engine()
        return CharmeMiddleware.__osEngine

    def process_request(self, request):
        USAGE_LOG.info(webusage(request))
        if request.method == 'OPTIONS':
            return HttpResponse(status=200)

        if CharmeMiddleware.get_store() is None:
            try:
                self.__init_store()
            except AttributeError, ex:
                messages.add_message(request, messages.ERROR, ex)
                messages.add_message(request, messages.INFO,
                                     'Missing configuration')
                return mm_render_to_response_error(request, '503.html', 503)

        if CharmeMiddleware.get_osengine() is None:
            try:
                self.__init_os_engine()
            except Exception, ex:
                messages.add_message(request, messages.ERROR, ex)
                messages.add_message(request, messages.INFO,
                                     'Missing configuration. '
                                     'Cannot initialize OpenSearch Engine')
                return mm_render_to_response_error(request, '503.html', 503)

        self._validate_request(request)

    def process_response(self, request, response):
        response['Access-Control-Allow-Origin'
            ] = request.META.get('HTTP_ORIGIN', request.build_absolute_uri())

        response['Access-Control-Allow-Credentials'] = 'true'

        response['Access-Control-Expose-Headers'] = (
            'Location, Content-Type, Content-Length')

        if request.method == 'OPTIONS':
            # Take default settings from class variable if no settings
            if getattr(settings, 'OPTIONS_HDR_RESPONSE'):
                for key, value in settings.OPTIONS_HDR_RESPONSE.items():
                    response[key] = value
            else:
                for key, value in (
                    self.__class__.DEFAULT_OPTIONS_HDR_RESPONSE.items()):
                    response[key] = value

            return response
        else:
            return response

    def process_exception(self, request, exception):
        LOGGING.info("process_exception - %s", str(exception))

    def _get_user_roles(self, user):
        # user.roles = contact_role_server
        # return user
        pass

    def _is_authenticated(self, request):
        pass

    def _validate_request(self, request):
        user = self._is_authenticated(request)
        # Here should compare:
        # - the request method ('get/put/post/delete')
        # - the request resource ('URI')
        pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('provider', '__first__'),
    ]

    operations = [
        migrations.CreateModel(
            name='FollowedResource',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('resource', models.CharField(help_text=b'The URI of the resource followed by the user.', max_length=500)),
                ('user', models.ForeignKey(help_text=b'The user to associate with this resource.', to=settings.AUTH_USER_MODEL)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Organization',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(help_text=b'The name of the organization.', unique=True, max_length=100)),
                ('primary_email', models.EmailField(help_text=b'The email of a person/list who is responsible for the CHAMRe client at this organization.', max_length=75)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='OrganizationClient',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('client', models.ForeignKey(to='provider.Client', help_text=b'The client to associate with this organization.', unique=True)),
                ('organization', models.ForeignKey(help_text=b'The name of the organization.', to='djcharme.Organization')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='OrganizationUser',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('role', models.CharField(help_text=b"The user's role within this organization.", max_length=10, choices=[(b'user', b'user'), (b'admin', b'admin')])),
                ('following', models.BooleanField(default=False, help_text=b'Send emails about any annotation created or altered by the organization.', verbose_name=b'email changes')),
                ('organization', models.ForeignKey(help_text=b'The name of the organization.', to='djcharme.Organization')),
                ('user', models.ForeignKey(help_text=b'The user to associate with this organization.', to=settings.AUTH_USER_MODEL)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='UserProfile',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('show_email', models.BooleanField(default=False, help_text=b'Include your email address in any new annotations you create.', verbose_name=b'show email')),
                ('user', models.OneToOneField(to=settings.AUTH_USER_MODEL)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



'''
BSD Licence
Copyright (c) 2015, Science & Technology Facilities Council (STFC)
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
        this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
        this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.
    * Neither the name of the Science & Technology Facilities Council (STFC)
        nor the names of its contributors may be used to endorse or promote
        products derived from this software without specific prior written
        permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Created on 18 Nov 2013

@author: mnagni
'''
import csv
from datetime import datetime
import logging

from django.conf import settings
from django.utils.http import urlquote
from rdflib import Literal
from rdflib.plugins.parsers.notation3 import BadSyntax

from djcharme import get_resource
from djcharme.exception import ParseError
from djcharme.node.actions import insert_rdf
from djcharme.node.constants import SUBMITTED
from djcharme.node.model_queries import get_client, get_user


LOGGING = logging.getLogger(__name__)

CITATION_TEMPLATE = '''
@prefix chnode: <http://localhost/> .
@prefix oa: <http://www.w3.org/ns/oa#> .
@prefix cito: <http://purl.org/spar/cito/> .
@prefix fabio: <http://purl.org/spar/fabio/> .
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix prov: <http://www.w3.org/ns/prov#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<chnode:annoID> a oa:Annotation, cito:CitationAct ;
    oa:hasTarget <load_target> ;
    oa:motivatedBy oa:linking ;
    oa:serializedAt "load_serialized_at"^^xsd:dateTime ;
    oa:serializedBy <chnode:agentID> ;
    cito:hasCitingEntity <load_body> ;
    cito:hasCitedEntity <load_target> ;
    cito:hasCitationCharacterization cito:citesAsDataSource .

<chnode:agentID> a prov:SoftwareAgent ;
    foaf:name "CHARMe system" .

<load_body> a fabio:load_classes .

<load_target> a fabio:MetadataDocument .
'''


def __load_datasets():
    """
    Load sample data sets.

    """
    datasets_file = open(get_resource('dataset_to_ceda_mappings.csv'))
    # datasets_file = open('resources/dataset_to_ceda_mappings.csv')
    dataset_reader = csv.reader(datasets_file, dialect='excel-tab')
    datasets = {}
    for row in dataset_reader:
        if type(row) != list \
            or len(row[0]) == 0 \
            or (len(row[1]) + len(row[2])) == 0 \
            or row[0] == 'Dataset':
            continue
        try:
            datasets[row[0]] = row[1:3]
        except Exception:
            pass
    return datasets


def __load_citations():
    """
    Load sample citations.

    """
    citations_file = open(get_resource
                          ('ceda_citations_to_metadata_url_mappings.csv'))
    citations_reader = csv.reader(citations_file, dialect='excel-tab')
    citations = {}
    dataset_name = None
    for row in citations_reader:
        if type(row) != list \
            or len(row[0]) == 0 \
            or (len(row[1]) + len(row[2])) == 0 \
            or row[0] == 'Dataset':
            continue
        try:
            if dataset_name == row[0]:
                citations[row[0]].append(row[1:])
            else:
                citations[row[0]] = [row[1:]]
            dataset_name = row[0]
        except Exception:
            pass
    return citations


def load_sample():
    """
    Load sample data.

    """
    LOGGING.debug('Loading samples')
    timestamp = Literal(datetime.utcnow())
    datasets = __load_datasets()
    citations = __load_citations()
    user_name = getattr(settings, 'SAMPLE_USER_NAME', 'stfc')
    user = get_user(user_name)
    client_name = getattr(settings, 'SAMPLE_CLIENT_NAME', 'STFC')
    client = get_client(client_name)
    serialized_at = timestamp
    if user == None or client == None:
        LOGGING.warn(('Samples not loaded as use %s not found and/or client for %s not found' % (user, client)))
        return
    for ds_key in datasets.keys():
        data_set = datasets.get(ds_key)
        cts = citations.get(ds_key, None)
        if cts == None:  # remove
            continue  # remove
        for cit in cts:
            annotation = CITATION_TEMPLATE.replace("load_target",
                                                   urlquote(data_set[1].strip()))
            # in press causes a
            # Exception: "file:///home/wilsona/workspace-space/in press" does not look like a valid URI, I cannot serialize this as N3/Turtle. Perhaps you wanted to urlencode it?
            if cit[8] and cit[8] != 'in press':
                annotation = annotation.replace("load_body", urlquote(cit[8].strip()))
            else:
                continue
            annotation = annotation.replace("load_serialized_at", serialized_at)
            if cit[0] == 'article':
                annotation = annotation.replace("load_classes", "Article")
            elif cit[0] == 'inbook':
                annotation = annotation.replace("load_classes", "BookChapter")
            elif cit[0] == 'proceedings':
                annotation = annotation.replace("load_classes",
                                                "AcademicProceedings")
            elif cit[0] == 'techreport':
                annotation = annotation.replace("load_classes",
                                                "TechnicalReport")
            elif cit[0] == 'misc':
                continue
            else:
                LOGGING.debug('Samples - found type %, ignoring record' % cit[0])
                continue

            try:
                insert_rdf(annotation, 'turtle', user, client,
                                   graph=SUBMITTED)
            except BadSyntax as ex:
                LOGGING.warn(ex)
                continue
            except ParseError as ex:
                LOGGING.warn(ex)
                continue
            except Exception as ex:
                LOGGING.error(ex)




# -*- coding: utf-8 -*-
from distutils.core import setup
from setuptools import find_packages
import re
import os

BASE_NAME = 'djcharme'

V_FILE = open(os.path.join(os.path.dirname(__file__),
                           BASE_NAME, '__init__.py'))

README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()

VERSION = re.compile(r".*__version__ = '(.*?)'",
                     re.S).match(V_FILE.read()).group(1)

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name=BASE_NAME,
    version=VERSION,
    author=u'Antony Wilson',
    author_email='antony.wilson@stfc.ac.uk',
    include_package_data=True,
    packages=find_packages(),  # include all packages under this directory
    url='https://github.com/CHARMe-Project/djcharme',
    license='BSD licence, see LICENCE',
    description='CHARMe Node',
    long_description=README,
    zip_safe=False,

    classifiers=[
        'Framework :: Django :: 1.7',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
    ],

    # Adds dependencies
    # django-registration==0.8 is explicitly included as 1.0 breaks
    # django-authopenid
    install_requires=[
        'ceda-markup==0.1.0',
        'cedatheme_mf54==1.0.0',
        'Django==1.7',
        'django-authopenid==1.0.2',
        'django-classy-tags==0.6.2',
        'django-cookie-law==1.0.6',
        'django-oauth2-provider==0.2.7.dev0',
        'django-registration==0.8',
        'html5lib==0.9999999',
        'isodate==0.5.4',
        'ordereddict==1.1',
        'py-bcrypt==0.4',
        'pyparsing==2.0.3',
        'rdflib==4.1.2',
        'rdflib-jsonld==0.2',
        'shortuuid==0.4.2',
        'SPARQLWrapper==1.6.4',
    ],
)




'''
BSD Licence
Copyright (c) 2014, Science & Technology Facilities Council (STFC)
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
        this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
        this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.
    * Neither the name of the Science & Technology Facilities Council (STFC)
        nor the names of its contributors may be used to endorse or promote
        products derived from this software without specific prior written
        permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Created on 25 May 2013

@author: Maurizio Nagni
'''
import datetime
import logging

from ceda_markup.atom.atom import createID, createUpdated, createPublished, \
    createEntry
from ceda_markup.atom.info import createContent, createTitle, TEXT_TYPE
from ceda_markup.opensearch import filter_results
from ceda_markup.opensearch.os_param import OSParam
from ceda_markup.opensearch.os_request import OS_NAMESPACE
from ceda_markup.opensearch.osquery import OSQuery
from ceda_markup.opensearch.template.atom import OSAtomResponse
from ceda_markup.opensearch.template.osresponse import OSEngineResponse, Result

from djcharme.node.constants import OA, CH_NODE, FOAF, RDF, RDFS, STABLE
from djcharme.node.search import annotation_resource, get_suggestions, \
    get_search_results
from djcharme.views import check_mime_format


LOGGING = logging.getLogger(__name__)

COUNT_DEFAULT = 10
START_INDEX_DEFAULT = 1
START_PAGE_DEFAULT = 1

ANNOTATIONS = "annotations"
SEARCH_TERMS = "searchTerms"


def _generate_url_id(url, context):
    if context.get('suggest') != None and context.get('suggest'):
        return "%s/suggest" % (url)
    else:
        return "%s/search" % (url)


def _import_count_and_page(context):
    ret = []

    try:
        ret.append(int(context.get('count')))
    except (ValueError, TypeError):
        ret.append(COUNT_DEFAULT)

    try:
        ret.append(int(context.get('startIndex')))
    except (ValueError, TypeError):
        ret.append(START_INDEX_DEFAULT)

    try:
        ret.append(int(context.get('startPage')))
    except (ValueError, TypeError):
        ret.append(START_PAGE_DEFAULT)

    return tuple(ret)


class COSAtomResponse(OSAtomResponse):
    """
    Class docs.

    """

    def __init__(self):
        """
        Constructor.

        """
        super(COSAtomResponse, self).__init__()

    def generate_entries(self, atomroot, subresults, path, \
                         params_model, context):
        LOGGING.debug("COSAtomResponse:generate_entries(atomroot, subresults, "\
                      "path, params_model, context)")
        if subresults is None:
            return

        entries = []
        for subresult in subresults:
            # Here could loop over results
            atom_id = createID(subresult['subject'], root=atomroot)
            ititle = createTitle(root=atomroot,
                                 body=subresult['subject'],
                                 itype=TEXT_TYPE)
            atom_content = createContent(root=atomroot,
                                         body=subresult['triple'],
                                         itype=TEXT_TYPE)
            time_doc = datetime.datetime.now().isoformat()
            atom_updated = createUpdated(time_doc,
                                         root=atomroot)
            atom_published = createPublished(time_doc,
                                             root=atomroot)
            entry = createEntry(atom_id, ititle, atom_updated,
                                published=atom_published,
                                content=atom_content, root=atomroot)

            entries.append(entry)

        for entry in entries:
            atomroot.append(entry)

    def generate_url(self, osHostURL, context):
        """
        Returns the proper URL to assemble the OSResponse links.

        """
        LOGGING.debug("COSAtomResponse:generate_url(%s, context)",
                      str(osHostURL))
        return _generate_url_id(osHostURL, context)

    def digest_search_results(self, results, context):
        LOGGING.debug("COSAtomResponse:digest_search_results(results, context)")
        title = "CHARMe results"
        count, start_index, start_page = _import_count_and_page(context)

        # set_subresults = set(results.subjects())
        # subjects = [subj for subj in set_subresults]
        # annotation_subresults = filter_results(results,
        #                                    count, start_index, start_page)

        iformat = context.get('format', 'json-ld')
        iformat = check_mime_format(iformat)
        if iformat == None:
            iformat = 'json-ld'
        if iformat == 'rdf':
            iformat = 'xml'
        elif iformat == 'ttl':
            iformat = 'turtle'
        if results['type'] == ANNOTATIONS:
            subresults = self._get_annotations(results['results'], iformat)
        elif results['type'] == SEARCH_TERMS:
            subresults = self._get_search_term_results(results['results'],
                                                       iformat)
        return Result(count, start_index, start_page, results['count'],
                      subresult=subresults, title=title)

    def _get_annotations(self, results, iformat):
        """

        """

        subresults = []
        for annotation_graph in results:
            try:
                subject = ([subj for subj in
                            annotation_graph.triples(annotation_resource())]
                           [0][0])
                subresults.append(
                    {'subject': str(subject),
                     'triple': annotation_graph.serialize(format=iformat)})
            except IndexError:
                LOGGING.warn("No Annotation resource for graph %s",
                             str(annotation_graph.serialize()))
                continue
        return subresults

    def _get_search_term_results(self, results, iformat):
        subresults = []
        for result in results:
            graph = result['graph']
            try:
                subresults.append(
                    {'subject': result['searchTerm'],
                     'triple': graph.serialize(format=iformat)})
            except IndexError:
                LOGGING.warn("No Annotation resource for graph %s",
                             str(graph.serialize()))
                continue
        return subresults

'''
    def generate_response(self, results, query, \
                          ospath, params_model, context):
        return results
'''


class COSRDFResponse(OSEngineResponse):
    """
    Class docs.

    """

    def __init__(self):
        """
        Constructor.

        """
        super(COSRDFResponse, self).__init__('rdf')

    def digest_search_results(self, results, context):
        LOGGING.debug("COSRDFResponse:digest_search_results(results, context)")
        title = "CHARMe results"
        count, start_index, start_page = _import_count_and_page(context)
        subresults = filter_results(results, count, start_index, start_page)
        return Result(count, start_index, start_page, len(results), \
                      subresult=subresults, title=title)
        # return results.serialize(format='xml')

    def generate_response(self, results, query, ospath, params_model, context):
        LOGGING.debug("COSRDFResponse:generate_response(results, query, " \
                      "ospath, params_model, context)")
        return results


class COSJsonLDResponse(OSEngineResponse):
    """
    Class docs.

    """

    def __init__(self):
        """
        Constructor.

        """
        super(COSJsonLDResponse, self).__init__('json-ld')

    def digest_search_results(self, results, context):
        LOGGING.debug("COSJsonLDResponse:digest_search_results(results, " \
                      "context)")
        return results.serialize(format='json-ld')

    def generate_response(self, results, query, ospath, params_model, context):
        LOGGING.debug("COSJsonLDResponse:generate_response(results, query, " \
                      "ospath, params_model, context)")
        return results


class COSTurtleResponse(OSEngineResponse):
    """
    Class docs.

    """

    def __init__(self):
        """
        Constructor.

        """
        super(COSTurtleResponse, self).__init__('ttl')

    def digest_search_results(self, results, context):
        LOGGING.debug("COSTurtleResponse:digest_search_results(results, " \
                      "context)")
        return results.serialize(format='turtle')

    def generate_response(self, results, query, ospath, params_model, context):
        LOGGING.debug("COSTurtleResponse:generate_response(results, query, " \
                      "ospath, params_model, context)")
        return results


class COSHTMLResponse(OSAtomResponse):
    """
    Class docs.

    """

    def __init__(self):
        """
        Constructor.

        """
        super(COSHTMLResponse, self).__init__()

    def generateResponse(self, result, queries, ospath, **kwargs):
        LOGGING.debug("COSHTMLResponse:generateResponse(result, queries, " \
                      "ospath, **kwargs)")
        return result + " HTML!"


class COSQuery(OSQuery):
    """
    Class docs.

    """

    def __init__(self):
        """
        Constructor.

        """
        params = []
        params.append(OSParam("count", "count",
                              namespace=OS_NAMESPACE,
                              default=str(COUNT_DEFAULT)))
        params.append(OSParam("startPage", "startPage",
                              namespace=OS_NAMESPACE,
                              default=str(START_PAGE_DEFAULT)))
        params.append(OSParam("startIndex", "startIndex",
                              namespace=OS_NAMESPACE,
                              default=str(START_INDEX_DEFAULT)))
        params.append(OSParam("q", "searchTerms",
                              namespace=OS_NAMESPACE, default=''))
        params.append(OSParam("bodyType", "type",
                              namespace=RDF, default=''))
        params.append(OSParam("citingType", "Resource",
                              namespace=RDFS, default=''))
        params.append(OSParam("comment", "chars",
                              namespace="http://www.w3.org/2011/content#",
                              default=''))
        params.append(OSParam("creatorFamilyName", "familyName",
                              namespace=FOAF, default=''))
        params.append(OSParam("creatorGivenName", "givenName",
                              namespace=FOAF, default=''))
        params.append(OSParam("dataType", "Resource",
                              namespace=RDFS, default=''))
        params.append(OSParam("domainOfInterest", "SemanticTag",
                              namespace=OA, default=''))
        params.append(OSParam("motivation", "Motivation",
                              namespace=OA, default=''))
        params.append(OSParam("organization", "name",
                              namespace=FOAF, default=''))
        params.append(OSParam("status", "status",
                              namespace=CH_NODE, default=STABLE))
        params.append(OSParam("target", "Resource",
                              namespace=RDFS, default=''))
        params.append(OSParam("title", "title",
                              namespace="http://purl.org/dc/terms/",
                              default=''))
        params.append(OSParam("userName", "accountName",
                              namespace=FOAF, default=''))
        params.append(OSParam("depth", "depth",
                              namespace=CH_NODE, default='1'))
        params.append(OSParam("format", "format",
                              namespace=CH_NODE, default='json-ld'))
        params.append(OSParam("sortKeys", "sortKeys",
                              namespace=CH_NODE, default='annotatedAt,DESC'))
        '''
        params.append(OSParam(BBOX, 'box',
                namespace = "http://a9.com/-/opensearch/extensions/geo/1.0/"))
        params.append(OSParam("start", "start",
                namespace = "http://a9.com/-/opensearch/extensions/time/1.0/"))
        params.append(OSParam("stop", "end",
                namespace = "http://a9.com/-/opensearch/extensions/time/1.0/"))
        '''
        self._query_signature = self._querySignature(params)
        super(COSQuery, self).__init__(params)

    def do_search(self, query, context):
        LOGGING.debug("do_search(query, context)")
        results = []
        total_results = 0
# TODO Hack until ceda_marckup supports suggest
        try:
            context.keys().index('suggest')
            if (query.attrib.get('q', None) != None
                and len(query.attrib.get('q')) > 0):
                results, total_results = get_suggestions(query.attrib['q'],
                                                         query.attrib)
            search_type = SEARCH_TERMS
        except ValueError:
            results, total_results = get_search_results(query.attrib)
            search_type = ANNOTATIONS
        return {'results':results, 'count':total_results, 'type':search_type}

    def do_suggest(self, query, context):
        LOGGING.debug("do_suggest(query, context)")
        results = None
        total_results = 0
        if query.attrib.get('q', None) != None \
                and len(query.attrib.get('q')) > 0:
            results, total_results = get_suggestions(query.attrib['q'],
                                                  query.attrib)
        # else: TODO error?
        return {'results':results, 'count':total_results, 'type':SEARCH_TERMS}

    def _querySignature(self, params_model):
        _params = []
        for params in params_model:
            if params.par_name not in ['count', 'startPage', 'startIndex']:
                _params.append(params.par_name)
        return _params




'''
BSD Licence
Copyright (c) 2014, Science & Technology Facilities Council (STFC)
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
        this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
        this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.
    * Neither the name of the Science & Technology Facilities Council (STFC)
        nor the names of its contributors may be used to endorse or promote
        products derived from this software without specific prior written
        permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Created on 24 Sep 2013

@author: mnagni
'''
import logging
import re
import string
from urlparse import urlparse

from rdflib.graph import Graph
from rdflib.namespace import RDF
from rdflib.term import Literal
from rdflib.term import URIRef

from djcharme.charme_middleware import CharmeMiddleware
from djcharme.node.triple_queries import extract_subject
from djcharme.node.actions import generate_graph
from djcharme.node.constants import STABLE


LOGGING = logging.getLogger(__name__)

PREFIX = """
PREFIX text: <http://jena.apache.org/text#>
PREFIX dcterm:  <http://purl.org/dc/terms/>
PREFIX oa: <http://www.w3.org/ns/oa#>
PREFIX cito: <http://purl.org/spar/cito/>
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
PREFIX cnt: <http://www.w3.org/2011/content#>
PREFIX foaf: <http://xmlns.com/foaf/0.1/>"""

ANNOTATIONS_SELECT = """
SELECT Distinct ?anno ?annotatedAt"""

ANNOTATIONS_SELECT_COUNT = """
SELECT (count(DISTINCT(?anno)) as ?count)"""

ANNOTATIONS_FOR_BODY_TYPE = """
?anno oa:annotatedAt ?annotatedAt .
?anno oa:hasBody ?body .
?body rdf:type ?bodyType ."""

ANNOTATIONS_FOR_CITING_TYPE = """
?anno oa:annotatedAt ?annotatedAt .
?anno cito:hasCitingEntity ?citingEntity .
?citingEntity rdf:type ?citingType ."""

ANNOTATIONS_FOR_COMMENT = """
?body text:query (cnt:chars '%s' 100) .
?anno oa:hasBody ?body ;
      oa:annotatedAt ?annotatedAt ."""

ANNOTATIONS_FOR_CREATOR_FAMILY_NAME = """
?anno oa:annotatedAt ?annotatedAt .
?anno oa:annotatedBy ?person .
?person foaf:familyName ?creatorFamilyName"""

ANNOTATIONS_FOR_CREATOR_GIVEN_NAME = """
?anno oa:annotatedAt ?annotatedAt .
?anno oa:annotatedBy ?person .
?person foaf:givenName ?creatorGivenName"""
      
ANNOTATIONS_FOR_DATA_TYPE = """
?anno oa:annotatedAt ?annotatedAt .
?anno oa:hasTarget ?target .
?target rdf:type ?dataType .
OPTIONAL {?target oa:item ?item .
?item rdf:type ?itemType}"""

ANNOTATIONS_FOR_DOMAIN = """
?anno oa:annotatedAt ?annotatedAt .
?anno oa:hasBody ?domainOfInterest .
?domainOfInterest rdf:type oa:SemanticTag ."""

ANNOTATIONS_FOR_MOTIVATION = """
?anno oa:annotatedAt ?annotatedAt .
?anno oa:motivatedBy ?motivation"""

ANNOTATIONS_FOR_ORGANIZATION = """
?anno oa:annotatedAt ?annotatedAt .
?anno oa:annotatedBy ?organization .
?organization rdf:type foaf:Organization .
?organization foaf:name ?organizationName ."""

ANNOTATIONS_FOR_REGION = """
"""  # TODO

ANNOTATIONS_FOR_STATUS = """
?anno oa:annotatedAt ?annotatedAt .
?anno oa:hasTarget ?target ."""

ANNOTATIONS_FOR_TARGET = """
?anno oa:annotatedAt ?annotatedAt .
?anno oa:hasTarget ?target .
OPTIONAL {?target oa:item ?item}"""

ANNOTATIONS_FOR_TITLE = """
?paper text:query (dcterm:title '%s' 100) .
?anno cito:hasCitingEntity ?paper ;
      oa:annotatedAt ?annotatedAt ."""

ANNOTATIONS_FOR_USER = """
?anno oa:annotatedAt ?annotatedAt .
?anno oa:annotatedBy ?person .
?person foaf:accountName ?userName"""

ANNOTATION_CLAUSES = {'bodyType':ANNOTATIONS_FOR_BODY_TYPE,
                      'citingType':ANNOTATIONS_FOR_CITING_TYPE,
                      'comment':ANNOTATIONS_FOR_COMMENT,
                      'creatorFamilyName':ANNOTATIONS_FOR_CREATOR_FAMILY_NAME,
                      'creatorGivenName':ANNOTATIONS_FOR_CREATOR_GIVEN_NAME,
                      'dataType':ANNOTATIONS_FOR_DATA_TYPE,
                      'domainOfInterest':ANNOTATIONS_FOR_DOMAIN,
                      'motivation':ANNOTATIONS_FOR_MOTIVATION,
                      'organization':ANNOTATIONS_FOR_ORGANIZATION,
                      'region':ANNOTATIONS_FOR_REGION,
                      'status':ANNOTATIONS_FOR_STATUS,
                      'target':ANNOTATIONS_FOR_TARGET,
                      'title':ANNOTATIONS_FOR_TITLE,
                      'userName':ANNOTATIONS_FOR_USER}

# The allowed values for the sort keys.
SORT_KEYS = ['bodyType',
             'citingType',
             'creatorFamilyName',
             'creatorGivenName',
             'dataType',
             'domainOfInterest',
             'motivation',
             'organization',
             'target',
             'userName',
             'annotatedAt'
             ]
             

def get_multi_value_parameter_names():
    """
    Get the list parameter names that can be used in an open searches and can
    have multiple occurrences.

    Returns:
        a list of parameter names

    """
    parameter_names = ANNOTATION_CLAUSES.keys()
    parameter_names.remove('status')
    return parameter_names


def annotation_resource(anno_uri=None):
    """
    Get an annotation subject, predicate, object using the anno_uri as the
    subject.

    Args:
        anno_uri (str): The annotation uri.

    Returns:
        str, str, str. Subject, predicate, object. If anno_uri is None then the
        returned subject will be None.

    """
    anno_ref = None
    if anno_uri:
        anno_ref = URIRef(anno_uri)
    return (anno_ref, RDF.type, URIRef('http://www.w3.org/ns/oa#Annotation'))


def _do__open_search(query_attr, graph, triples):
    depth = int(query_attr.get('depth', 3))
    ret = _populate_annotations(graph, triples, depth)
    return ret


def _populate_annotations(graph, triples, depth=3):
    ret = []
    for row in triples:
        tmp_g = Graph()
        for subj in extract_subject(graph, row[0], depth):
            tmp_g.add(subj)
        ret.append(tmp_g)
    return ret


def _populate_graph(graph, triples):
    tmp_g = Graph()
    for row in triples:
        for subj in _extract_subject_form_graph(graph, row[0]):
            tmp_g.add(subj)
    return tmp_g


def _extract_subject_form_graph(graph, subject):
    tmp_g = Graph()
    for res in graph.triples((subject, None, None)):
        tmp_g.add(res)
    return tmp_g


def _get_count(triples):
    """
    Retrieve the value of a count query from a list of triples.

    Args:
        triples ([triple]): The results from a SELECT count(?) query

    Returns:
        an int of the count
    """
    for triple in triples:
        return int(triple[0])


def _get_limit(query_attr):
    """
    Get the limit for the number of results to return.

    Args:
        query_attr (dict): The query parameters from the users request.

    """
    limit = int(query_attr.get('count'))
    return limit


def _get_offset(query_attr):
    """
    Get the off set of results to return.

    Args:
        query_attr (dict): The query parameters from the users request.

    """
    limit = _get_limit(query_attr)
    LOGGING.debug("Using limit: %s", limit)
    offset = (int(query_attr.get('startPage', 1)) - 1) * limit
    offset = offset + int(query_attr.get('startIndex', 1)) - 1
    if offset > 0:
        LOGGING.debug("Using offset: %s", offset)
        return " OFFSET " + str(offset)
    else:
        return ""


def _get_graph(query_attr):
    """
    Get the graph based on the value of the attribute 'status'.

    If the users has not set a value for 'status' the use the STABLE graph

    Args:
        query_attr (dict): The query parameters from the users request.
    """
    graph_name = str(query_attr.get('status', STABLE))
    return generate_graph(CharmeMiddleware.get_store(), graph_name)


class SearchProxy(object):
    def __init__(self, query):
        _query = query
        self.query_signature = None
        super(SearchProxy, self).__init__(self)


def get_suggestions(parameter_names, query_attr):
    """
    Get the values for the given parameter names.


    Args:
        parameter_names (str): The parameter names to search for. This should be
            a space separated list or an '*'. For a list of available parameter
            names use get_parameter_names().
        query_attr (dict): The query parameters from the users request.

    Returns:
        list of dict. Key words:
            graph - A graph containing the results
            count - The count of available results
            searchTerm - The parameter name that was searched for

    """
    LOGGING.debug("get_suggestions(%s, query_attr)", parameter_names)
    graph = _get_graph(query_attr)
    limit = _get_limit(query_attr)
    offset = _get_offset(query_attr)
    results = []
    total_results = 0
    parameter_name_list = (re.sub('[' + string.punctuation + ']', '',
                                  parameter_names).split())
    if parameter_names == "*":
        parameter_name_list = ["*"]
    where_clause = ''
    for parameter_name in ANNOTATION_CLAUSES.keys():
        where_clause = (where_clause +
                        _get_where_for_parameter_name(query_attr,
                                                      parameter_name))
    for parameter_name in parameter_name_list:
        if parameter_name == "bodyType" or parameter_name == "*":
            result, count = _get_body_types(graph, "bodyType", where_clause,
                                            limit, offset)
            results.append(result)
            total_results = total_results + count
        if parameter_name == "citingType" or parameter_name == "*":
            result, count = _get_citing_types(graph, "citingType", where_clause,
                                            limit, offset)
            results.append(result)
            total_results = total_results + count
        if parameter_name == "dataType" or parameter_name == "*":
            result, count = _get_data_types(graph, "dataType", where_clause,
                                            limit, offset)
            results.append(result)
            total_results = total_results + count
        if parameter_name == "domainOfInterest" or parameter_name == "*":
            result, count = _get_domains_of_interest(graph, "domainOfInterest",
                                                     where_clause, limit,
                                                     offset)
            results.append(result)
            total_results = total_results + count
        if parameter_name == "motivation" or parameter_name == "*":
            result, count = _get_motivations(graph, "motivation", where_clause,
                                             limit, offset)
            results.append(result)
            total_results = total_results + count
        if parameter_name == "organization" or parameter_name == "*":
            result, count = _get_organizations(graph, "organization",
                                               where_clause, limit, offset)
            results.append(result)
            total_results = total_results + count
    return results, total_results


def _get_body_types(graph, parameter_name, where_clause, limit, offset):
    statement = (PREFIX +
    """
    SELECT Distinct ?bodyType
    WHERE {
    {%s}
    ?anno oa:hasBody ?body .
    ?body rdf:type ?bodyType .
    }
    ORDER BY ?bodyType
    LIMIT %s
    %s""" % (where_clause, limit, offset))
    count_statement = (PREFIX +
    """
    SELECT  count (Distinct ?bodyType)
    WHERE {
    {%s}
    ?anno oa:hasBody ?body .
    ?body rdf:type ?bodyType .
    }""") % where_clause
    return _do_query(graph, parameter_name, statement, count_statement,
                     "http://www.w3.org/2004/02/skos/core#prefLabel")


def _get_citing_types(graph, parameter_name, where_clause, limit, offset):
    statement = (PREFIX +
    """
    SELECT Distinct ?citingType
    WHERE {
    {%s}
    ?anno oa:annotatedAt ?annotatedAt .
    ?anno cito:hasCitingEntity ?citingEntity .
    ?citingEntity rdf:type ?citingType .
    }
    ORDER BY ?citingType
    LIMIT %s
    %s""" % (where_clause, limit, offset))
    count_statement = (PREFIX +
    """
    SELECT  count (Distinct ?citingType)
    WHERE {
    {%s}
    ?anno oa:annotatedAt ?annotatedAt .
    ?anno cito:hasCitingEntity ?citingEntity .
    ?citingEntity rdf:type ?citingType .
    }""") % where_clause
    return _do_query(graph, parameter_name, statement, count_statement,
                     "http://www.w3.org/2004/02/skos/core#prefLabel")


def _get_data_types(graph, parameter_name, where_clause, limit, offset):
    statement = (PREFIX +
    """
    SELECT Distinct ?dataType
    WHERE {
    {%s}
    ?anno oa:hasTarget ?target .
    ?target rdf:type ?dataType .
    FILTER (?dataType != oa:Composite)
    }
    """ % (where_clause))
    result1, _ = _do_query(graph, parameter_name, statement, None,
                     "http://www.w3.org/2004/02/skos/core#prefLabel")

    statement = (PREFIX +
    """
    SELECT Distinct ?itemType
    WHERE {
    {%s}
    ?anno oa:hasTarget ?target .
    ?target oa:item ?item .
    ?item rdf:type ?itemType .
    }
    ORDER BY ?itemType
    """ % (where_clause))
    result2, _ = _do_query(graph, parameter_name, statement, None,
                     "http://www.w3.org/2004/02/skos/core#prefLabel")

    result = {'searchTerm': parameter_name}
    # Combine the graphs
    tmp_graph = result1['graph']
    for res in result2['graph']:
        tmp_graph.add(res)

    count = len(tmp_graph)
    result['count'] = count

    # filter on limit and offset
    statement = (PREFIX +
    """
    SELECT ?s ?p ?o
    WHERE {
    ?s ?p ?o
    }
    ORDER BY ?s
    LIMIT %s
    %s""" % (limit, offset))
    triples = tmp_graph.query(statement)
    result['graph'] = Graph()
    for res in triples:
        result['graph'].add(res)
    LOGGING.debug("_get_data_types returning %s triples out of %s for %s",
                  len(result['graph']), result['count'], parameter_name)
    return result, count


def _get_domains_of_interest(graph, parameter_name, where_clause, limit,
                             offset):
    statement = (PREFIX +
    """
    SELECT DISTINCT ?body ?domainOfInterest
    WHERE {
    {%s}
    ?anno oa:hasBody ?body .
    ?body rdf:type oa:SemanticTag .
    ?body skos:prefLabel ?domainOfInterest .
    }
    ORDER BY ?domainOfInterest
    LIMIT %s
    %s""" % (where_clause, limit, offset))
    count_statement = (PREFIX +
    """
    SELECT count (Distinct ?domainOfInterest)
    WHERE {
    {%s}
    ?anno oa:hasBody ?body .
    ?body rdf:type oa:SemanticTag .
    ?body skos:prefLabel ?domainOfInterest .
    }""") % where_clause
    return _do_query(graph, parameter_name, statement, count_statement,
                     "http://www.w3.org/2004/02/skos/core#prefLabel")


def _get_motivations(graph, parameter_name, where_clause, limit, offset):
    statement = (PREFIX +
    """
    SELECT Distinct ?motivation
    WHERE {
    {%s}
    ?anno oa:motivatedBy ?motivation .
    }
    ORDER BY ?motivation
    LIMIT %s
    %s""" % (where_clause, limit, offset))
    count_statement = (PREFIX +
    """
    SELECT count (Distinct ?motivation)
    WHERE {
    {%s}
    ?anno oa:motivatedBy ?motivation .
    }""") % where_clause
    return _do_query(graph, parameter_name, statement, count_statement,
                     "http://www.w3.org/2004/02/skos/core#prefLabel")


def _get_organizations(graph, parameter_name, where_clause, limit, offset):
    statement = (PREFIX +
    """
    SELECT DISTINCT ?organization ?name
    WHERE {
    {%s}
    ?anno oa:annotatedBy ?organization .
    ?organization rdf:type foaf:Organization .
    ?organization foaf:name ?name .
    }
    ORDER BY ?name
    LIMIT %s
    %s""" % (where_clause, limit, offset))
    count_statement = (PREFIX +
    """
    SELECT count (Distinct ?name)
    WHERE {
    {%s}
    ?anno oa:annotatedBy ?organization .
    ?organization rdf:type foaf:Organization .
    ?organization foaf:name ?name .
    }""") % where_clause
    return _do_query(graph, parameter_name, statement, count_statement,
                     "http://xmlns.com/foaf/0.1/name")


def _do_query(graph, parameter_name, statement, count_statement,
              labelType):
    LOGGING.debug("_do_query %s", statement)
    result = {'searchTerm': parameter_name}
    if count_statement != None:
        result['count'] = _get_count(graph.query(count_statement))
    else:
        result['count'] = -1
    triples = graph.query(statement)
    result['graph'] = Graph()
    url = urlparse(labelType)
    obj = ''
    for row in triples:
        if len(row) > 1:
            obj = row[1]
        else:
            url = urlparse(row[0])
            obj = ''
            if url.fragment != '':
                obj = url.fragment
            elif url.path != None:
                path = str(url.path)
                bits = path.split('/')
                obj = bits[(len(bits) - 1)]
            else:
                obj = row[0]
        result['graph'].add((URIRef(row[0]), URIRef(labelType), Literal(obj)))
    LOGGING.debug("_do_query returning %s triples out of %s for %s",
                  len(result['graph']), result['count'], parameter_name)
    return result, result['count']


def get_search_results(query_attr):
    """
    Get the annotations which refer to the given parameters.

    Args:
        query_attr (dict): The query parameters from the users request.

    Returns:
        list of graphs. The result of the search.

    """
    LOGGING.debug("get_search_results(query_attr)")
    where_clause = _get_where_clause(query_attr)
    annotations_order = _get_annotations_order(query_attr)

    statement = PREFIX + ANNOTATIONS_SELECT + where_clause + annotations_order
    if _get_limit(query_attr) > 0:
        statement = statement + ' LIMIT ' + str(_get_limit(query_attr))
    if _get_offset(query_attr) != None:
        statement = statement + _get_offset(query_attr)

    statement_count = PREFIX + ANNOTATIONS_SELECT_COUNT + where_clause
    graph = _get_graph(query_attr)
    triples = graph.query(statement)
    results = _do__open_search(query_attr, graph, triples)
    total_results = _get_count(graph.query(statement_count))
    LOGGING.debug("get_search_results returning %s triples out of %s",
                  len(results), total_results)
    return results, total_results


def _get_where_clause(query_attr):
    """
    Get the where clause based on the values in query_attr.

    Args:
        query_attr (dict): The query parameters from the users request.

    Returns:
        str The where clause.

    """
    where_clause = ''
    parameter_names = ANNOTATION_CLAUSES.keys()
    unused_parameter_names = ANNOTATION_CLAUSES.keys()
    for parameter_name in parameter_names:
        clause_bit = _get_where_for_parameter_name(query_attr,
                                                      parameter_name)
        if len(clause_bit) > 0 :
            where_clause = where_clause + clause_bit
            unused_parameter_names.remove(parameter_name)
    
    # we may need extra bits if there is a sortKeys in the parameters so that
    # we can use it later
    sort_keys = _get_sort_key_values(query_attr)
    for sort_name, _ in sort_keys:
        if sort_name in unused_parameter_names:
            clause = 'OPTIONAL{' + ANNOTATION_CLAUSES[sort_name] + '}'
            where_clause = where_clause + clause

    # special case for status
    if len(where_clause) == 0:
        where_clause = where_clause = '{' + ANNOTATION_CLAUSES['status'] + '}'
    where_clause = ' WHERE {' + where_clause + '}'
    return where_clause


def _get_where_for_parameter_name(query_attr, parameter_name):
    """
    Add the where clause and filter for the given parameter name.

    Args:
        query_attr (dict): The query parameters from the users request.
        parameter_name (str): The name of the parameter.

    Returns:
        str The where clause.
    """
    values = query_attr.get(parameter_name)
    if values == None or len(values) < 1 or parameter_name == 'status':
        return  ''
    if parameter_name == 'title' or parameter_name == 'comment':
        return '{' + (ANNOTATION_CLAUSES[parameter_name] % values) + '}'

    # TODO remove split when I can work out to handle the parameter being
    # returned as a list.
    values = values.split()
    where_clause = '{' + ANNOTATION_CLAUSES[parameter_name] + ' FILTER ('
    first_term = True
    for value in values:
        if first_term:
            first_term = False
        else:
            where_clause = where_clause + ' || '
        if (parameter_name == 'creatorFamilyName'
            or parameter_name == 'creatorGivenName'):
            # case insensitive search on given and family names
            where_clause = (where_clause + 'regex(?' + parameter_name + ', "^'
                            + value + '$", "i")')
        else:
            where_clause = where_clause + '?' + parameter_name + "="
            if (parameter_name == 'userName'):
                where_clause = where_clause + "'" + value + "'"
            else:
                where_clause = where_clause + "<" + URIRef(value) + '>'
            # special case for composite
            if parameter_name == 'target':
                where_clause = where_clause + ' || '
                where_clause = where_clause + '?' + "item="
                where_clause = where_clause + "<" + URIRef(value) + '>'
            # special case for composite
            if parameter_name == 'dataType':
                where_clause = where_clause + ' || '
                where_clause = where_clause + '?' + "itemType="
                where_clause = where_clause + "<" + URIRef(value) + '>'
    where_clause = where_clause + ')} '

    return where_clause


def _get_annotations_order(query_attr):
    """
    Validate the values for the sortKeys paramerter and constuct the order
    statment. Bad values will result in defaults being used.

    Args:
        query_attr (dict): The query parameters from the users request.

    Returns:
        str The order statment.
    """
    first = True
    sort_values = _get_sort_key_values(query_attr)
    for sort_name, order_name in sort_values:
        if first:
            first = False
            annotations_order = '\nORDER BY %s(?%s)' % (order_name, sort_name)
        else:
            annotations_order = '%s %s(?%s)' % (annotations_order, order_name,
                                                sort_name)
    return annotations_order


def _get_sort_key_values(query_attr):
    """
    Extract and validate the values from the sortKeys paramerter. If ASC/DESC
    is not provided then the default of DESC is used. The sortKeys must be a
    case sensitive match for a SORT_KEYS value otherwise they are ignored. Bad
    values will result in defaults being returned.

    Args:
        query_attr (dict): The query parameters from the users request.

    Returns:
        list[(str, str)] A list of (variable name, ASC/DESC).
    """
    values = []
    sortKeys = query_attr.get('sortKeys').split(' ')
    for sortKey in sortKeys:
        bits = sortKey.split(',')
        sort_attr = bits[0]
        if sort_attr in SORT_KEYS:
            if len(bits) > 1 and (bits[1].upper() == 'ASC' 
                                  or bits[1].upper() == 'DESC'):
                order = bits[1].upper()
            else:
                order = 'DESC'
            values.append((sort_attr, order))
    if len(values) == 0:
        values.append(('annotatedAt', 'DESC'))
    return values



'''
BSD Licence
Copyright (c) 2015, Science & Technology Facilities Council (STFC)
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
        this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
        this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.
    * Neither the name of the Science & Technology Facilities Council (STFC)
        nor the names of its contributors may be used to endorse or promote
        products derived from this software without specific prior written
        permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Created on 14 May 2013

@author: mnagni
'''
import json
import logging

from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.http.response import HttpResponseBadRequest, \
    HttpResponseRedirectBase, HttpResponse, HttpResponseNotFound, \
    HttpResponseForbidden, HttpResponseNotAllowed, HttpResponseServerError
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import View

from djcharme import mm_render_to_response_error, \
    __version__
from djcharme.exception import NotFoundError
from djcharme.exception import ParseError
from djcharme.exception import SecurityError
from djcharme.exception import StoreConnectionError
from djcharme.exception import UserError
from djcharme.models import FollowedResource
from djcharme.node import is_following_resource, resource_exists
from djcharme.node.actions import collect_annotations, find_resource_by_id, \
    format_resource_uri_ref, change_annotation_state, get_vocab, \
    report_to_moderator, validate_graph_name
from djcharme.node.actions import insert_rdf, modify_rdf
from djcharme.node.constants import OA, FOAF, PROV, RDF, FORMAT_MAP, \
    CONTENT_HTML, CONTENT_JSON, CONTENT_RDF, CONTENT_TEXT, DATA, FOLLOWING, \
    PAGE, SUBMITTED, RETIRED
from djcharme.views import isPOST, isOPTIONS, \
    validate_mime_format, http_accept, get_depth, content_type, \
    check_mime_format, get_format
from djcharme.views.resource import agent, annotation, annotation_index, \
    activity, composite, person, resource


LOGGING = logging.getLogger(__name__)


class HttpResponseCreated(HttpResponse):
    """
    Implements a simple HTTP 201 response.
    """
    status_code = 201


class HttpResponseNoContent(HttpResponse):
    """
    Implements a simple HTTP 204 response.
    """
    status_code = 204


class HttpResponseSeeOther(HttpResponseRedirectBase):
    """
    Implements a simple HTTP 303 response.
    """
    status_code = 303



class HttpResponseNotAcceptable(HttpResponse):
    """
    Implements a simple HTTP 406 response.
    """        
    status_code = 406


class HttpResponseUnsupportedMediaType(HttpResponse):
    """
    Implements a simple HTTP 415 response.
    """        
    status_code = 415


def __serialize(graph, req_format=CONTENT_RDF):
    '''
        Serializes a graph according to the required format
        - rdflib:Graph **graph** the graph to serialize
        - string **req_format** the serialization format
        - **return** the serialized graph
    '''
    if req_format == FORMAT_MAP['json-ld']:
        req_format = 'json-ld'

    return graph.serialize(format=req_format)


def index(request, graph='submitted'):
    """
    Returns a tabular view of the stored annotations.
    - HTTPRequest **request** the client request
    - string **graph**  the required named graph
    TODO: In a future implementation this actions should be supported by an
    OpenSearch implementation

    """
    try:
        validate_graph_name(graph)
    except UserError as ex:
        if CONTENT_HTML in http_accept(request):
            messages.add_message(request, messages.ERROR, str(ex))
            return mm_render_to_response_error(request, '400.html', 400)
        else:
            return HttpResponseBadRequest(str(ex))
    tmp_g = None
    try:
        tmp_g = collect_annotations(graph)
    except StoreConnectionError as ex:
        LOGGING.error("index - unexpected error: %s", ex)
        if CONTENT_HTML in http_accept(request):
            messages.add_message(request, messages.ERROR, str(ex))
            return mm_render_to_response_error(request, '500.html', 500)
        else:
            return HttpResponseServerError(str(ex))

    req_format = validate_mime_format(request)
    if req_format is not None:
        return HttpResponse(__serialize(tmp_g, req_format=req_format))
    elif CONTENT_HTML in http_accept(request):
        return annotation_index(request, tmp_g, graph)

    return HttpResponseNotAcceptable('Format not accepted')


@csrf_exempt
def insert(request):
    """
    Inserts in the triplestore a new annotation under the "SUBMITTED" graph

    """
    try:
        return _insert(request)
    except UserError as ex:
        return HttpResponseBadRequest(str(ex))
    except Exception as ex:
        LOGGING.error("insert - unexpected error: %s", ex)
        return HttpResponseServerError(str(ex))


def _insert(request):
    """
    Inserts in the triplestore a new annotation under the "SUBMITTED" graph

    """
    request_format = check_mime_format(content_type(request))

    if request_format is None:
        return HttpResponseUnsupportedMediaType("Cannot ingest the posted format")

    if isPOST(request) or isOPTIONS(request):
        triples = request.body
        try:
            anno_uri = insert_rdf(triples, request_format, request.user,
                                  request.client, graph=SUBMITTED)
        except ParseError as ex:
            LOGGING.debug("insert parsing error: %s", str(ex))
            return HttpResponseBadRequest(str(ex))
        except StoreConnectionError as ex:
            LOGGING.error("insert - unexpected error: %s", ex)
            return HttpResponseServerError(str(ex))

        response = HttpResponseCreated(anno_uri, content_type=CONTENT_TEXT)
        response['Location'] = anno_uri
        return response
    
    return HttpResponseNotAllowed(["POST", "OPTIONS"],
                                  content='Only POST and OPTION methods are allowed')


@csrf_exempt
def modify(request):
    """
    Modify the annotation contained in the request.

    Args:
        request (WSGIRequest): The request from the user

    """
    try:
        return _modify(request)
    except NotFoundError as ex:
        return HttpResponseNotFound('Resource does not exist on this system')
    except SecurityError as ex:
        return HttpResponseForbidden(str(ex))
    except UserError as ex:
        return HttpResponseBadRequest(str(ex))
    except Exception as ex:
        LOGGING.error("modify - unexpected error: %s", str(ex))
        return HttpResponseServerError(str(ex))


def _modify(request):
    """
    Modify the annotation contained in the request.

    Args:
        request (WSGIRequest): The request from the user

    """
    request_format = check_mime_format(content_type(request))

    if request_format is None:
        return HttpResponseUnsupportedMediaType("Cannot ingest the posted format")

    if isPOST(request) or isOPTIONS(request):
        try:
            anno_uri = modify_rdf(request, request_format)
        except ParseError as ex:
            LOGGING.debug("modify parsing error: %s", str(ex))
            return HttpResponseBadRequest(str(ex))
        except StoreConnectionError as ex:
            LOGGING.error("modify - unexpected error: %s", str(ex))
            return HttpResponseServerError(str(ex))
        response = HttpResponseCreated(anno_uri, content_type=CONTENT_TEXT)
        response['Location'] = anno_uri
        return response

    return HttpResponseNotAllowed(["POST", "OPTIONS"],
                                  content='Only POST and OPTION methods are allowed')


@csrf_exempt
def advance_status(request):
    '''
        Advance the status of an annotation
    '''
    try:
        return _advance_status(request)
    except NotFoundError as ex:
        return HttpResponseNotFound('Resource does not exist on this system')
    except SecurityError as ex:
        return HttpResponseForbidden(str(ex))
    except UserError as ex:
        return HttpResponseBadRequest(str(ex))
    except Exception as ex:
        LOGGING.error("advance_status - unexpected error: %s", str(ex))
        return HttpResponseServerError(str(ex))


def _advance_status(request):
    '''
        Advance the status of an annotation
    '''
    if isPOST(request) and (CONTENT_JSON in content_type(request) or
                            'application/json' in content_type(request)):
        try:
            params = json.loads(request.body)
        except ValueError as ex:
            return HttpResponseBadRequest(str(ex))
        if not params.has_key('annotation') or not params.has_key('toState'):
            return HttpResponseBadRequest("Missing annotation/state parameters")
        LOGGING.info("advancing %s to state:%s", str(params.get('annotation')),
                     str(params.get('toState')))
        change_annotation_state(params.get('annotation'),
                                params.get('toState'), request)
        return HttpResponseNoContent()
    elif not isPOST(request):
        return HttpResponseNotAllowed(["POST"],
                                      content='Only POST method is allowed')
    else:
        return HttpResponseUnsupportedMediaType("Message must contain %s" % 
                                                CONTENT_JSON)
    return HttpResponseNoContent()


class Following(View):
    """
    This class enables a user to follow or stop following a resource and to
    list the resources that they are following.

    """
    @method_decorator(csrf_exempt)
    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        try:
            return super(Following, self).dispatch(*args, **kwargs)
        except Exception as ex:
            LOGGING.error("Following - unexpected error: %s", ex)
            return HttpResponseServerError(str(ex))

    def delete(self, request, *args, **kwargs):
        """
        Stop following the resource.
    
        """
        resource_uri = kwargs["resource_uri"]
        FollowedResource.objects.filter(
            user=self.request.user).filter(resource=resource_uri).delete()
        return HttpResponseNoContent()

    def get(self, request, *args, **kwargs):
        """
        Get a list of resources that the user is following.
        
        """
        # if text format redirect to GUI
        if CONTENT_HTML in http_accept(request):
            LOGGING.info("Redirecting to /following/")
            path = "/%s/" % FOLLOWING
            return HttpResponseSeeOther(path)
        if CONTENT_JSON not in http_accept(request):
            return HttpResponseNotAcceptable('Format not accepted')
        following = (FollowedResource.objects.filter(user=request.user)
                     .order_by('resource'))
        data = []
        for follow in following:
            data.append({"resource": follow.resource})
        return HttpResponse(json.dumps(data), content_type=CONTENT_JSON)

    def put(self, request, *args, **kwargs):
        """
        Validate the resource. Check that the user is not already following the
        resource and that the resource exists in the triple store. Add a new
        record to the FollowedResource model.
        
        """
        resource_uri = kwargs["resource_uri"]
        if is_following_resource(request.user, resource_uri):
            # You are already following this resource
            return HttpResponseNoContent()
        if not (resource_exists(resource_uri)):
            return HttpResponseNotFound('Resource does not exist on this system')
    
        followed_resource = FollowedResource.objects.create(
                                user_id=request.user.id,
                                resource=resource_uri)
        followed_resource.save()
        return HttpResponseNoContent()
 

class ReportToModerator(View):
    """
    Report the resource to the moderator.

    """
    @method_decorator(csrf_exempt)
    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        try:
            return super(ReportToModerator, self).dispatch(*args, **kwargs)
        except Exception as ex:
            LOGGING.error("ReportToModerator - unexpected error: %s", ex)
            return HttpResponseServerError(str(ex))

    def put(self, request, *args, **kwargs):
        resource_id = kwargs["resource_id"]
        LOGGING.info("reporting %s to moderator", resource_id)
        c_type = content_type(request)
        if c_type is not None and CONTENT_TEXT not in c_type:
            return HttpResponseUnsupportedMediaType("Message must contain %s" % 
                                                    CONTENT_TEXT)
        try:
            report_to_moderator(request, resource_id)
        except NotFoundError as ex:
            return HttpResponseNotFound(str(ex))
        except SecurityError as ex:
            return HttpResponseForbidden(str(ex))
        return HttpResponseNoContent()
    
    
class Resource(View):
    """
    Get a resources details or delete the resource.

    """
    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        try:
            return super(Resource, self).dispatch(*args, **kwargs)
        except Exception as ex:
            LOGGING.error("Resource - unexpected error: %s", ex)
            return HttpResponseServerError(str(ex))

    def delete(self, request, *args, **kwargs):
        """
        Delete the resource, move it to the 'retired' graph.

        """
        resource_id = kwargs["resource_id"]
        LOGGING.info("advancing %s to state:%s", resource_id, RETIRED)
        try:
            change_annotation_state(resource_id, RETIRED, request)
        except NotFoundError as ex:
            return HttpResponseNotFound(str(ex))
        except SecurityError as ex:
            return HttpResponseForbidden(str(ex))
        return HttpResponseNoContent()

    def get(self, request, *args, **kwargs):
        """
        Get the resource details dependent on the mime format.

        """
        resource_id = kwargs["resource_id"]
        # if valid format requested redirect to DATA
        if validate_mime_format(request) is not None:
            path = "/%s/%s" % (DATA, resource_id)
            path = self.process_resource_parameters(request, path)
            LOGGING.info("Redirecting to %s", path)
            return HttpResponseSeeOther(path)

        # if text format redirect to PAGE
        if CONTENT_HTML in http_accept(request):
            path = '/%s/%s' % (PAGE, resource_id)
            path = self.process_resource_parameters(request, path)
            LOGGING.info("Redirecting to %s", path)
            return HttpResponseSeeOther(path)

        return HttpResponseNotAcceptable('Format not accepted')


    def process_resource_parameters(self, request, path):
        """
        Add depth and format parameters onto the path.

        """
        depth = get_depth(request)
        format_ = get_format(request)
        if format_ is not None:
            path = "%s/?format=%s" % (path, format_)
            if depth is not None:
                path = "%s&depth=%s" % (path, depth)
        elif depth is not None:
            path = "%s/?depth=%s" % (path, depth)
        return path

    
class ResourceData(View):
    """
    Get the resources details.

    """
    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        try:
            return super(ResourceData, self).dispatch(*args, **kwargs)
        except Exception as ex:
            LOGGING.error("ResourceData - unexpected error: %s", ex)
            return HttpResponseServerError(str(ex))

    def get(self, request, *args, **kwargs):
        """
        Get the resource details dependent on the mime format.

        """
        resource_id = kwargs["resource_id"]
        req_format = validate_mime_format(request)
        if req_format is None:
            return HttpResponseNotAcceptable('Format not accepted')

        depth = get_depth(request)
        if depth == None:
            depth = 1
        tmp_g = find_resource_by_id(resource_id, depth)
        if len(tmp_g) < 1:
            return HttpResponseNotFound('Resource does not exist on this system')
        return HttpResponse(tmp_g.serialize(format=req_format, auto_compact=False),
                            content_type=FORMAT_MAP.get(req_format))


class ResourcePage(View):
    """
    Get the resources details and present them in a html page.

    """
    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        try:
            return super(ResourcePage, self).dispatch(*args, **kwargs)
        except Exception as ex:
            LOGGING.error("ResourcePage - unexpected error: %s", ex)
            messages.add_message(self.request, messages.ERROR, str(ex))
            return mm_render_to_response_error(self.request, '500.html', 500)

    def get(self, request, *args, **kwargs):
        """
        Get the resources details and present them in a html page.

        """
        resource_id = kwargs["resource_id"]
        if CONTENT_HTML not in http_accept(request):
            return HttpResponseNotAcceptable('Format not accepted')

        tmp_g = find_resource_by_id(resource_id, 1)
        if len(tmp_g) < 1:
            messages.add_message(self.request, messages.ERROR,
                                 'Resource does not exist on this system')
            return mm_render_to_response_error(self.request, '404.html', 404)
        resource_uri = format_resource_uri_ref(resource_id)
    
        # Check if the resource is an annotation
        triples = tmp_g.triples((resource_uri, RDF['type'], OA['Annotation']))
        for _ in triples:
            return annotation(request, resource_uri, tmp_g)

        # Check if the resource is a SoftwareAgent
        triples = tmp_g.triples((resource_uri, RDF['type'],
                                 PROV['SoftwareAgent']))
        for _ in triples:
            return agent(request, resource_uri, tmp_g)
    
        # Check if the resource is an activity
        triples = tmp_g.triples((resource_uri, RDF['type'], PROV['Activity']))
        for _ in triples:
            return activity(request, resource_uri, tmp_g)
    
        # Check if the resource is a person
        triples = tmp_g.triples((resource_uri, RDF['type'], FOAF['Person']))
        for _ in triples:
            return person(request, resource_uri, tmp_g)
    
        # Check if the resource is a composite
        triples = tmp_g.triples((resource_uri, RDF['type'], OA['Composite']))
        for _ in triples:
            return composite(request, resource_uri, tmp_g)
        return resource(request, resource_uri, tmp_g)


def version(request):
    """
    Get the version number of the server.

    Args:
        request (WSGIRequest): The request from the user

    """
    return HttpResponse(__version__, content_type=CONTENT_TEXT)


def vocab(request):
    """
    Get the chame vocab.

    Args:
        request (WSGIRequest): The request from the user

    """
    tmp_g = None
    try:
        tmp_g = get_vocab()
    except StoreConnectionError as ex:
        LOGGING.error("vocab - unexpected error: %s", ex)
        return HttpResponseServerError(str(ex))

    req_format = validate_mime_format(request)
    if req_format is None or CONTENT_HTML in http_accept(request):
        req_format = CONTENT_JSON
    return HttpResponse(__serialize(tmp_g, req_format=req_format))


